﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Syncfusion.Windows.Controls.Grid;
using Syncfusion.Grouping;
using System.Collections;
using System.ComponentModel;
using Syncfusion.Windows.Controls.Cells;
using System.Windows.Media;
using System.Windows;
using System.Windows.Controls;
using Syncfusion.Windows.GridCommon;
using Syncfusion.Windows.Styles;
using System.Windows.Media.Imaging;

namespace GroupingSample
{
    /// <summary>
    /// This class adds grouping support to a GridControl. To use this class, you must have a GridControl and 
    /// an IEnumerable class that serves as the datasource. Properties are exposed to give you access to all
    /// visuals related the display of the grid.
    /// </summary>
    public class GroupingHelper
    {
        #region initialization

        /// <summary>
        /// 
        /// </summary>
        /// <param name="gridControl">The GridControl</param>
        /// <param name="dataSource">The IEnumerable datasource.</param>
        public GroupingHelper(GridControl gridControl, IEnumerable dataSource)
        {
            this.gridControl = gridControl;
            this.groupingEngine = new LimitedEngine(this); //new Engine();// new LimitedEngine(this);
            // Metrics.DisplayElapsedTime("start setting DataSource");
            this.groupingEngine.SetSourceList(dataSource);
            //  Metrics.DisplayElapsedTime("end setting DataSource start InitHelper");
            InitHelper();
            //  Metrics.DisplayElapsedTime("end InitHelper");
        }

        /// <summary>
        /// A placeholder name for the column of row headers.
        /// </summary>
        public string ColumnName_RowHeader = "_ROWHEADERCOLUMN";

        /// <summary>
        /// A placeholder name for the expand column.
        /// </summary>
        public string ColumnName_Expand = "_EXPANDCOLUMN";

        private void LoadDefaultStyles()
        {
            ColumnStyles.Clear();

            ColumnStyleInfo csi = new ColumnStyleInfo();
            csi.Name = ColumnName_RowHeader;
            ColumnStyles.Add(ColumnName_RowHeader, csi);

            csi = new ColumnStyleInfo();
            csi.Name = ColumnName_Expand;
            ColumnStyles.Add(ColumnName_Expand, csi);

            foreach (FieldDescriptor fd in GroupingEngine.TableDescriptor.Fields)
            {
                csi = new ColumnStyleInfo();
                csi.Name = fd.MappingName;
                ColumnStyles.Add(fd.MappingName, csi);
            }
        }

        private void InitHelper()
        {
            ResetRowColumnCount();
            WireEvents();

            GridControl.ColumnWidths[expandColumn] = 2 * ExpandIndentWidth;
            GridControl.ColumnWidths[rowHeaderColumn] = Metrics.Enabled ? GridControl.ColumnWidths.DefaultLineSize : 30;

            LoadDefaultStyles();


            GridControl.Model.CellModels.Add("SortHeader", new GridCellModel<SortHeaderCellRenderer>());
            SortHeaderCellRenderer renderer = GridControl.CellRenderers["SortHeader"] as SortHeaderCellRenderer;
            renderer.InitializeRenderer(GridControl, this);

            GridControl.Model.Options.AllowSelection &= ~GridSelectionFlags.Column;
            GridControl.Model.Options.ActivateCurrentCellBehavior = GridCellActivateAction.ClickOnCell;

            GridControl.ContextMenu = Menu;

        }

        /// <summary>
        /// This method subscribes to GridControl and Grouping Engine events to implement grouping support on the GridControl.
        /// </summary>
        public virtual void WireEvents()
        {
            if (GridControl != null)
            {
                GridControl.Model.QueryCellInfo += new GridQueryCellInfoEventHandler(Model_QueryCellInfo);
                GridControl.Model.CommitCellInfo += new GridCommitCellInfoEventHandler(Model_CommitCellInfo);
                GridControl.Model.QueryCoveredRange += new GridQueryCoveredRangeEventHandler(Model_QueryCoveredRange);
                GridControl.PreviewMouseLeftButtonDown += new System.Windows.Input.MouseButtonEventHandler(GridControl_PreviewMouseLeftButtonDown);
                GridControl.PreviewMouseRightButtonDown += new System.Windows.Input.MouseButtonEventHandler(GridControl_PreviewMouseRightButtonDown);
                GridControl.PreviewMouseRightButtonUp += new System.Windows.Input.MouseButtonEventHandler(GridControl_PreviewMouseRightButtonUp);
                GridControl.CurrentCellMoved += new GridCurrentCellMovedEventHandler(GridControl_CurrentCellMoved);
                gridControl.CurrentCellStartEditing += new CancelEventHandler(gridControl_CurrentCellStartEditing);

            }
            if (GroupingEngine != null)
            {
                GroupingEngine.SourceListListChangedCompleted += new TableListChangedEventHandler(GroupingEngine_SourceListListChangedCompleted);
                GroupingEngine.TableDescriptor.GroupedColumns.Changed += new Syncfusion.Collections.ListPropertyChangedEventHandler(GroupedColumns_Changed);
                GroupingEngine.GroupSummaryInvalidated += new GroupEventHandler(GroupingEngine_GroupSummaryInvalidated);
            }
        }

        public virtual void UnwireEvents()
        {
            if (GridControl != null)
            {
                GridControl.Model.QueryCellInfo -= new GridQueryCellInfoEventHandler(Model_QueryCellInfo);
                GridControl.Model.CommitCellInfo -= new GridCommitCellInfoEventHandler(Model_CommitCellInfo);
                GridControl.Model.QueryCoveredRange -= new GridQueryCoveredRangeEventHandler(Model_QueryCoveredRange);
                GridControl.PreviewMouseLeftButtonDown -= new System.Windows.Input.MouseButtonEventHandler(GridControl_PreviewMouseLeftButtonDown);
                GridControl.PreviewMouseRightButtonDown -= new System.Windows.Input.MouseButtonEventHandler(GridControl_PreviewMouseRightButtonDown);
                GridControl.CurrentCellMoved -= new GridCurrentCellMovedEventHandler(GridControl_CurrentCellMoved);
                gridControl.CurrentCellStartEditing -= new CancelEventHandler(gridControl_CurrentCellStartEditing);

            }
            if (GroupingEngine != null)
            {
                GroupingEngine.SourceListListChangedCompleted -= new TableListChangedEventHandler(GroupingEngine_SourceListListChangedCompleted);
                GroupingEngine.TableDescriptor.GroupedColumns.Changed -= new Syncfusion.Collections.ListPropertyChangedEventHandler(GroupedColumns_Changed);
                gridControl.CurrentCellStartEditing -= new CancelEventHandler(gridControl_CurrentCellStartEditing);
            }
        }


        private void ToggleGroupColumn(int columnIndex)
        {
            if (columnIndex >= fixedColumns)
            {
                string name = ColumnIndexToName(columnIndex);
                if (name != null && name.Length > 0)
                {
                    ToggleGroupColumn(name);
                }
            }
        }

        /// <summary>
        /// Groups or ungroups a given column.
        /// </summary>
        /// <param name="name">The name of the column.</param>
        public void ToggleGroupColumn(string name)
        {
            ClearCurrentCellAndSelections();
            bool isGrouped = GroupingEngine.TableDescriptor.GroupedColumns.IndexOf(name) > -1;
            if (isGrouped)
            {
                GroupingEngine.TableDescriptor.GroupedColumns.Remove(name);
            }
            else
            {
                GroupingEngine.TableDescriptor.GroupedColumns.Add(name);
            }
        }

        /// <summary>
        /// Expands all groups.
        /// </summary>
        public void ExpandAllGroups()
        {
            ClearCurrentCellAndSelections();
            GroupingEngine.Table.ExpandAllGroups();
            ResetDisplay();
        }

        /// <summary>
        /// Collapses all groups.
        /// </summary>
        public void CollapseAllGroups()
        {
            ClearCurrentCellAndSelections();
            GroupingEngine.Table.CollapseAllGroups();
            if (IsTopLevelGroupHidden())
            {
                //need to reopen top group, other wise nothing visible
                GroupingEngine.Table.TopLevelGroup.IsExpanded = true;
            }
            ResetDisplay();
        }

        private bool IsTopLevelGroupHidden()
        {
            int count = 0;
            return GridControl.RowHeights.GetHidden(rowTopLevelCaption, out count);
        }
        /// <summary>
        /// Call the method after changing any operation that changes the number of rows and columns in the grid.
        /// </summary>
        public void ResetDisplay()
        {
            ResetDisplay(true);
        }

        /// <summary>
        /// Call the method after changing any operation that changes the number of rows and columns in the grid.
        /// </summary>
        /// <param name="resetCellSpans">True if the RecordCellSpanBackground collection should be recreated.</param>
        public void ResetDisplay(bool resetCellSpans)
        {
            Metrics.DisplayNote("ResetDisplay");

            if (resetCellSpans && RecordSpanInfoBrush != null)
            {
                GridControl.CellSpanBackgrounds.Clear();
                RemoveRecordCellSpanBackgrounds(GroupingEngine.Table.TopLevelGroup);
            }

            GridControl.UnloadArrangedCells();
            GridControl.RenderedCellVisuals.RemoveRows(0, int.MaxValue, null);
            ResetRowColumnCount();

            if (GridControl.CurrentCell.IsInMoveTo && GridControl.CurrentCell.MoveFromRowIndex > GridControl.Model.RowCount)
            {

            }
            if (resetCellSpans)
            {
                if (RecordSpanInfoBrush != null)
                    PopulateRecordCellSpanBackgrounds(GroupingEngine.Table.TopLevelGroup);
                if (SpacerSpanInfoBrush != null)
                    PopulateSpacerCellSpanBackgrounds(GroupingEngine.Table.TopLevelGroup);
            }
            GridControl.InvalidateVisual(true);
        }

        /// <summary>
        /// Override this methos to control the string displayed in a Caption.
        /// </summary>
        /// <param name="cr">The CaptionRow whose display text is to be returned.</param>
        /// <returns>The text that is displayed in the caption.</returns>
        public virtual string GetCaptionDisplayText(CaptionRow cr)
        {
            // return cr.ParentGroup.CategoriesToString();
            return cr.ParentGroup.IsTopLevelGroup ? string.Format("{0} items", GroupingEngine.Table.UnsortedRecords.Count)
                : string.Format("[{0}]={1}", cr.ParentGroup.Name, cr.ParentGroup.Category);
        }

        private Dictionary<string, string> summaryPropertyName1;

        /// <summary>
        /// Holds a list of column names with their associated summary display property.
        /// </summary>
        internal Dictionary<string, string> SummaryPropertyName
        {
            get
            {
                if (summaryPropertyName1 == null)
                {
                    summaryPropertyName1 = new Dictionary<string, string>();
                }
                return summaryPropertyName1;
            }
            set { summaryPropertyName1 = value; }
        }

        /// <summary>
        /// Sets a summary.
        /// </summary>
        /// <param name="columnName">The name of the column whose summary is being set.</param>
        /// <param name="sd">The SumamryDescriptor defining the aggregate calculation.</param>
        /// <param name="resetDisplay">Whther or not to immediately reset the grid display after setting this summary.</param>
        /// <param name="summaryPropertyName">The property name from teh aggregate calculation to be displayed.</param>
        public void SetSummary(string columnName, SummaryDescriptor sd, bool resetDisplay, string summaryPropertyName)
        {
            int i = -1;
            sd.Name = columnName; //change the defaultname to match the column...
            if ((i = GroupingEngine.TableDescriptor.Summaries.IndexOf(columnName)) > -1)
            {
                if (sd != null)
                {
                    GroupingEngine.TableDescriptor.Summaries[i] = sd;
                }
                else
                {
                    SummaryPropertyName.Remove(columnName);
                    GroupingEngine.TableDescriptor.Summaries.Remove(columnName);
                }
            }
            else
            {
                if (sd != null)
                {
                    SummaryPropertyName.Add(columnName, summaryPropertyName);
                    GroupingEngine.TableDescriptor.Summaries.Add(sd);
                }
            }
            if (resetDisplay)
            {
                ResetDisplay();
            }
        }

        private void PopulateRecordCellSpanBackgrounds(Group g)
        {
            if (g != null && g.IsExpanded && g.Records != null && g.Records.Count > 0)
            {
                int topRowIndex = GroupingEngine.Table.DisplayElements.IndexOf(g.Caption) + 1 + (ShowCaption ? 1 : 0);
                int botRowIndex = topRowIndex + g.Records.Count - 1;
                GridControl.CellSpanBackgrounds.Add(new CellSpanBackgroundInfo(topRowIndex, fixedColumns, botRowIndex, GridControl.Model.ColumnCount, false, false, RecordSpanInfoBrush, null));
            }
            else if (g != null && g.IsExpanded && g.Groups != null && g.Groups.Count > 0)
            {
                foreach (Group g1 in g.Groups)
                {
                    PopulateRecordCellSpanBackgrounds(g1);
                }
            }
        }

        private void RemoveRecordCellSpanBackgrounds(Group g)
        {
            if (g != null)
            {
                int topRowIndex = GroupingEngine.Table.DisplayElements.IndexOf(g.Caption) + 1 + (ShowCaption ? 1 : 0);
                List<CellSpanBackgroundInfo> list = GridControl.CellSpanBackgrounds.GetCellSpanBackgrounds(topRowIndex, fixedColumns);
                if (list != null)
                {
                    foreach (CellSpanBackgroundInfo span in list)
                    {
                        GridControl.CellSpanBackgrounds.Remove(span);
                    }
                }
            }
        }

        private void PopulateSpacerCellSpanBackgrounds(Group g)
        {
            if (g != null && g.IsExpanded && g.Records != null && g.Records.Count > 0)
            {
                int topRowIndex = GroupingEngine.Table.DisplayElements.IndexOf(g.Caption) + 1 + (ShowCaption ? 1 : 0);
                int botRowIndex = topRowIndex + g.Records.Count - 1;
                GridControl.CellSpanBackgrounds.Add(new CellSpanBackgroundInfo(topRowIndex, expandColumn, botRowIndex, expandColumn, false, false, SpacerSpanInfoBrush, null));
            }
            else if (g != null && g.IsExpanded && g.Groups != null && g.Groups.Count > 0)
            {
                foreach (Group g1 in g.Groups)
                {
                    PopulateSpacerCellSpanBackgrounds(g1);
                }
            }
        }

        private void RemoveSpacerCellSpanBackgrounds(Group g)
        {
            if (g != null)
            {
                int topRowIndex = GroupingEngine.Table.DisplayElements.IndexOf(g.Caption) + 1 + (ShowCaption ? 1 : 0);
                List<CellSpanBackgroundInfo> list = GridControl.CellSpanBackgrounds.GetCellSpanBackgrounds(topRowIndex, expandColumn);
                if (list != null)
                {
                    foreach (CellSpanBackgroundInfo span in list)
                    {
                        GridControl.CellSpanBackgrounds.Remove(span);
                    }
                }
            }
        }

        private void ClearCurrentCellAndSelections()
        {
            GridControl.CurrentCell.Deactivate(true);
            GridControl.Model.Selections.Clear();
        }

        private void ResetRowColumnCount()
        {
            int newRowCount = GroupingEngine.Table.DisplayElements.Count + 1; // columnheader row + 1
            GridControl.Model.RowCount = newRowCount;
            GridControl.Model.ColumnCount = VisibleColumns.Count + fixedColumns; //GroupingEngine.TableDescriptor.Fields.Count + fixedColumns - 1;
        }


        #endregion

        #region GroupingEngine events
        void GroupingEngine_SourceListListChangedCompleted(object sender, TableListChangedEventArgs e)
        {
            Metrics.DisplayNote("SourceListListChangedCompleted: " + e.ListChangedType.ToString());
            switch (e.ListChangedType)
            {
                case ListChangedType.ItemAdded:
                case ListChangedType.ItemDeleted:
                    ResetRowColumnCount();
                    break;
                //case ListChangedType.ItemChanged:
                //    break;
                default:
                    break;
            }
        }

        private int GetExpandColumnWidth()
        {
            int width = 2 * ExpandIndentWidth;
            return width;
        }



        void GroupedColumns_Changed(object sender, Syncfusion.Collections.ListPropertyChangedEventArgs e)
        {
            Metrics.DisplayNote("GroupedColumns_Changed ");
            if (AllowGroupIndent)
            {
                GridControl.Model.ColumnWidths[expandColumn] = (GroupingEngine.TableDescriptor.GroupedColumns.Count + 2) * ExpandIndentWidth;
            }
            ResetDisplay();
        }

        void GroupingEngine_GroupSummaryInvalidated(object sender, GroupEventArgs e)
        {
            Metrics.DisplayNote("GroupSummaryInvalidated ");
            ResetDisplay();
        }

        #endregion

        #region GridControl events

        //used for sorting and expand/collapse support
        void GridControl_PreviewMouseLeftButtonDown(object sender, System.Windows.Input.MouseButtonEventArgs e)
        {
            GridControlBase grid = (GridControlBase)sender;
            RowColumnIndex cell = grid.PointToCellRowColumnIndex(e);

            if (AllowClickSort && cell.RowIndex == 0 && cell.ColumnIndex >= fixedColumns) //header row
            {
                string name = ColumnIndexToName(cell.ColumnIndex);
                int i = -1;
                if ((i = GroupingEngine.TableDescriptor.SortedColumns.IndexOf(name)) > -1)
                {
                    SortColumnDescriptor scd = GroupingEngine.TableDescriptor.SortedColumns[i];
                    scd.SortDirection = scd.SortDirection == ListSortDirection.Ascending ?
                        ListSortDirection.Descending : ListSortDirection.Ascending;
                }
                else
                {
                    GroupingEngine.TableDescriptor.SortedColumns.Clear();
                    GroupingEngine.TableDescriptor.SortedColumns.Add(name);
                }
                ResetDisplay(true);
                e.Handled = true;
            }
            else if (cell.ColumnIndex == expandColumn)
            {//expand collapse a group
                Element el;
                DisplayRowType drt = GetDisplayRowType(cell.RowIndex, out el);
                if (drt == DisplayRowType.Caption)
                {
                    Group g = el.ParentGroup;
                    if (RecordSpanInfoBrush != null)
                        RemoveRecordCellSpanBackgrounds(GroupingEngine.Table.TopLevelGroup);
                    if (SpacerSpanInfoBrush != null)
                        RemoveSpacerCellSpanBackgrounds(GroupingEngine.Table.TopLevelGroup);
                    g.IsExpanded = !g.IsExpanded;
                    ClearCurrentCellAndSelections();
                    ResetDisplay(true);
                    e.Handled = true;
                }
            }
        }


        void Model_QueryCoveredRange(object sender, GridQueryCoveredRangeEventArgs e)
        {
            if (e.CellRowColumnIndex.ColumnIndex >= fixedColumns && e.CellRowColumnIndex.RowIndex < GridControl.Model.RowCount)
            {
                Element el;
                DisplayRowType drt = GetDisplayRowType(e.CellRowColumnIndex.RowIndex, out el);
                if (drt == DisplayRowType.Caption)
                {
                    e.Range = new CoveredCellInfo(e.CellRowColumnIndex.RowIndex, fixedColumns, e.CellRowColumnIndex.RowIndex, GridControl.Model.ColumnCount - 1);
                    e.Handled = true;
                }
            }
        }

        void GridControl_CurrentCellMoved(object sender, GridCurrentCellMovedEventArgs e)
        {
            if (GroupingEngine.Table.AddNewRecord.IsEditing)
            {
                Element el;
                RowColumnIndex cell = GridControl.CurrentCell.CellRowColumnIndex;
                DisplayRowType drt = GetDisplayRowType(cell.RowIndex, out el);
                if (drt != DisplayRowType.AddNewRecord)
                {
                    GroupingEngine.Table.AddNewRecord.EndEdit();
                    RowColumnIndex save = GridControl.CurrentCell.CellRowColumnIndex;
                    ClearCurrentCellAndSelections();
                    ResetDisplay();
                    GridControl.CurrentCell.CellRowColumnIndex = save;
                }
                needtoResetDueToValueChanged = false;
            }
            else if (needtoResetDueToValueChanged)
            {
                needtoResetDueToValueChanged = false;
                //ClearCurrentCellAndSelections();
                ResetDisplay();
                if (RowIndexToPosition(GridControl.CurrentCell.RowIndex) >= GroupingEngine.Table.DisplayElements.Count)
                {
                    ClearCurrentCellAndSelections();
                }
            }

        }

        void gridControl_CurrentCellStartEditing(object sender, CancelEventArgs e)
        {
            Element el;
            RowColumnIndex cell = GridControl.CurrentCell.CellRowColumnIndex;
            DisplayRowType drt = GetDisplayRowType(cell.RowIndex, out el);
            if (drt == DisplayRowType.AddNewRecord)
            {
                if (!GroupingEngine.Table.AddNewRecord.IsEditing)
                {
                    // GroupingEngine.Table.AddNew();
                    GroupingEngine.Table.AddNewRecord.BeginEdit();
                }
            }
        }

        private bool needtoResetDueToValueChanged = false;
        void Model_CommitCellInfo(object sender, GridCommitCellInfoEventArgs e)
        {
            Element el;
            DisplayRowType drt = GetDisplayRowType(e.Cell.RowIndex, out el);
            needtoResetDueToValueChanged = false;

            switch (drt)
            {
                case DisplayRowType.AddNewRecord:
                case DisplayRowType.LastRecord:
                case DisplayRowType.Record:
                    if (e.Cell.ColumnIndex >= fixedColumns)
                    {
                        string columnName = ColumnIndexToName(e.Cell.ColumnIndex);
                        RecordRow rec = el as RecordRow;
                        if (rec != null)
                        {
                            rec.GetRecord().SetValue(columnName, e.Style.CellValue);
                            e.Handled = true;

                            if (GroupingEngine.TableDescriptor.SortedColumns.IndexOf(columnName) > -1
                                || GroupingEngine.TableDescriptor.GroupedColumns.IndexOf(columnName) > -1)
                            {
                                //ClearCurrentCellAndSelections();
                                //ResetDisplay();
                                needtoResetDueToValueChanged = true;
                            }
                        }
                    }
                    break;
                default:
                    break;
            }
        }

        void Model_QueryCellInfo(object sender, GridQueryCellInfoEventArgs e)
        {
            Element el;
            DisplayRowType drt = GetDisplayRowType(e.Cell.RowIndex, out el);
            switch (drt)
            {
                #region DisplayRowType.Caption

                case DisplayRowType.Caption:
                    if (e.Cell.ColumnIndex == fixedColumns)
                    {
                        //put caption text into cell
                        CaptionRow cr = el as CaptionRow;
                        if (cr != null)
                        {
                            e.Style.ModifyStyle(CellStyleForCaption, StyleModifyType.Override);
                            e.Style.Text = GetCaptionDisplayText(cr);
                            //if (ShiftCaptionByLevel > 0)
                            //    e.Style.BorderMargins.Left += (cr.ParentGroup.GroupLevel + 1) * ShiftCaptionByLevel;
                        }
                    }
                    else if (e.Cell.ColumnIndex == expandColumn)
                    {
                        e.Style.ModifyStyle(CellStyleForExpand, StyleModifyType.Override);

                        string s = el.ParentGroup.IsExpanded ? "-" : "+";
                        ////if (el.ParentGroup.GroupLevel > -1 && shiftCaptionByLevel > 0)
                        ////{
                        ////    s = new string(' ', 2 * (el.ParentGroup.GroupLevel + 1)) + s;
                        ////}
                        if (AllowGroupIndent)
                        {
                            int width = (el.ParentGroup.GroupLevel + 1) * ExpandIndentWidth + 5;
                            e.Style.BorderMargins.Left = width;

                        }
                        e.Style.Text = s;
                    }
                    else if (e.Cell.ColumnIndex == rowHeaderColumn)
                    {
                        e.Style.ModifyStyle(CellStyleForCaptionRowHeader, StyleModifyType.Override);
                    }
                    break;
                #endregion

                #region DisplayRowType.ColumnHeader

                case DisplayRowType.ColumnHeader:
                    if (e.Cell.ColumnIndex >= fixedColumns)
                    {
                        //set up headers
                        e.Style.ModifyStyle(CellStyleForColumnHeader, StyleModifyType.Override);

                        e.Style.Text = ColumnStyles[ColumnIndexToName(e.Cell.ColumnIndex)].HeaderText;
                    }
                    else if (e.Cell.ColumnIndex == rowHeaderColumn)
                    {
                        e.Style.ModifyStyle(CellStyleForColumnHeaderRowHeader, StyleModifyType.Override);
                    }
                    else if (e.Cell.ColumnIndex == expandColumn)
                    {
                        e.Style.ModifyStyle(CellStyleForColumnHeaderRowHeader, StyleModifyType.Override);
                    }
                    break;
                #endregion

                #region DisplayRowType.SummarySection

                case DisplayRowType.SummarySection:
                    if (e.Cell.ColumnIndex == expandColumn)
                    {
                        e.Style.ModifyStyle(CellStyleForSummarySpacer, StyleModifyType.Override);
                    }
                    else if (e.Cell.ColumnIndex == rowHeaderColumn)
                    {
                        e.Style.ModifyStyle(CellStyleForSummaryRowHeader, StyleModifyType.Override);
                    }
                    else // >= fixedColumns
                    {
                        //set the value first
                        Group g = el.ParentGroup;
                        string columnName = ColumnIndexToName(e.Cell.ColumnIndex);
                        if (GroupingEngine.TableDescriptor.Summaries.IndexOf(columnName) > -1)
                        {
                            e.Style.CellValue = g.GetSummaryProperty(columnName, SummaryPropertyName[columnName]);
                        }

                        //first apply column style properties...
                        if (ColumnStyles.ContainsKey(columnName))
                        {
                            e.Style.ModifyStyle(ColumnStyles[columnName].Style, StyleModifyType.Override);
                        }

                        //then apply specific SummaryCell style...
                        Pen p = e.Style.Borders.Right;
                        e.Style.ModifyStyle(CellStyleForSummary, StyleModifyType.Override);

                        if (CellStyleForSummary.Borders.Right == null &&
                            (e.Cell.ColumnIndex < fixedColumns ||
                             e.Cell.ColumnIndex == GridControl.Model.ColumnCount - 1)
                            )
                        {
                            e.Style.Borders.Right = p;
                        }
                    }
                    break;
                #endregion

                #region  DisplayRowType.LastRecord   DisplayRowType.Record   DisplayRowType.AddNewRecord

                case DisplayRowType.AddNewRecord:
                case DisplayRowType.LastRecord:
                case DisplayRowType.Record:
                    if (e.Cell.ColumnIndex >= fixedColumns)
                    {
                        string columnName = ColumnIndexToName(e.Cell.ColumnIndex);
                        RecordRow rec = el as RecordRow;
                        if (rec != null)
                        {
                            e.Style.CellValue = rec.GetRecord().GetValue(columnName);
                            if (ColumnStyles.ContainsKey(columnName))
                            {
                                e.Style.ModifyStyle(ColumnStyles[columnName].Style, StyleModifyType.Override);
                            }
                        }
                    }
                    else if (e.Cell.ColumnIndex == expandColumn)
                    {
                        if (drt == DisplayRowType.Record)
                        {
                            e.Style.ModifyStyle(CellStyleForSpacer, StyleModifyType.Override);
                        }
                        else  //DisplayRowType.LastRecord
                        {
                            e.Style.ModifyStyle(CellStyleForLastSpacer, StyleModifyType.Override);
                        }
                    }
                    else if (e.Cell.ColumnIndex == rowHeaderColumn)
                    {
                        if (drt == DisplayRowType.Record)
                        {
                            e.Style.ModifyStyle(CellStyleForRecordRowHeader, StyleModifyType.Override);
                        }
                        else if (drt == DisplayRowType.AddNewRecord)
                        {
                            e.Style.ModifyStyle(CellStyleForAddNewRowHeader, StyleModifyType.Override);
                        }
                        else //DisplayRowType.LastRecord
                        {
                            e.Style.ModifyStyle(CellStyleForLastRecordRowHeader, StyleModifyType.Override);
                        }
                    }
                    break;

                #endregion

                default:
                    break;
            }

            #region for debug
            if (Metrics.Enabled && e.Cell.RowIndex > 0 && e.Cell.ColumnIndex == 0)
            {
                // e.Style.Text = drt.ToString();
                //  Element el = GroupingEngine.Table.DisplayElements[e.Cell.RowIndex - 1];
                if (el != null)
                {
                    string s = el.GetType().ToString();
                    int loc = s.LastIndexOf('.');
                    e.Style.Text = drt.ToString();// s.Substring(loc + 1);
                }
            }
            #endregion
        }

        #endregion

        #region Helper Methods

        private int expandColumn
        {
            get { return fixedColumns - 1; }
        }
        private int fixedColumns
        {
            get { return GridControl.Model.HeaderColumns + 1; }
        }

        private int rowHeaderColumn
        {
            get { return fixedColumns - 2; }
        }
        private int rowTopLevelCaption
        {
            get { return 1; }
        }


        //TODO: see if changing this to a look up is better  - keep a dictonary of <element type, DisplayRowType> that is populated on demand...
        /// <summary>
        /// Returns the Grouping display element type of the object at the given RowIndex in the GridControl.
        /// </summary>
        /// <param name="rowIndex">The given RowIndex.</param>
        /// <param name="el">The Groupingdisplay Element at the given RowIndex.</param>
        /// <returns></returns>
        public DisplayRowType GetDisplayRowType(int rowIndex, out Element el)
        {
            DisplayRowType drt = DisplayRowType.Record;
            el = null;
            if (rowIndex == 0)
            {
                drt = DisplayRowType.ColumnHeader;
            }
            else
            {
                int position = RowIndexToPosition(rowIndex);
                IList list = GroupingEngine.Table.DisplayElements != null
                    ? GroupingEngine.Table.DisplayElements : GroupingEngine.Table.NestedDisplayElements;

                if (position < 0 || position >= list.Count)
                {
                    return DisplayRowType.Unknown;
                }

                el = list[position] as Element;
                if (el == null)
                {
                    return DisplayRowType.Unknown;
                }

                Type t = el.GetType();
                if (t == typeof(CaptionRow))
                {
                    drt = DisplayRowType.Caption;
                }
                else if (t == typeof(SummarySection))
                {
                    drt = DisplayRowType.SummarySection;
                }
                else
                {
                    if (el.GetRecord() is AddNewRecord)
                    {
                        drt = DisplayRowType.AddNewRecord;
                    }
                    //else if (rowIndex == list.Count - 1 ||
                    //        !(list[RowIndexToPosition(rowIndex + 1)] is RecordRow)
                    //     )
                    else if (rowIndex == list.Count ||
                            !(list[RowIndexToPosition(rowIndex + 1)] is RecordRow)
                         )
                    {
                        drt = DisplayRowType.LastRecord;
                    }
                }
            }
            return drt;
        }

        /// <summary>
        /// Returns the position in the Grouping DisplayeElements collection that corresponds to the RowIndex of the GridControl.
        /// </summary>
        /// <param name="rowIndex"></param>
        /// <returns></returns>
        public int RowIndexToPosition(int rowIndex)
        {
            return rowIndex - 1;// (IsTopLevelGroupHidden() ? 0 : 1); //-1 is captionrow
        }

        /// <summary>
        /// Returns the name of the field in the Grouping Engine that corresponds to the given ColumnIndex.
        /// </summary>
        /// <param name="columnIndex">The ColumnIndex in the GridControl.</param>
        /// <returns></returns>
        public string ColumnIndexToName(int columnIndex)
        {
            if (columnIndex - fixedColumns < 0)
            {
                string s = "";
                switch (columnIndex - fixedColumns)
                {
                    case -1:
                        s = ColumnName_Expand;
                        break;
                    case -2:
                        s = ColumnName_RowHeader;
                        break;
                    default:
                        break;
                }
                return s;
            }

            return VisibleColumns[columnIndex - fixedColumns];
        }

        #endregion

        #region menu stuff
        private ContextMenu menu;

        //must be in the order of menu 
        const int groupMenuItem = 0;
        const int showCaptionMenuItem = 1;
        const int showAddNewMenuItem = 2;
        const int showSummaryMenuItem = 3;
        const int showRowHeaderMenuItem = 4;
        const int showTopLevelGroupMenuItem = 5;
        const int expandAllMenuItem = 6;
        const int collapseAllMenuItem = 7;
        const int clearSortMenuItem = 8;
        const int clearAllGroupsMenuItem = 9;


        /// <summary>
        /// The Context menu associated with the GridControl that exposes support for grouping and showing.hiding certain
        /// GridControl display elements.
        /// </summary>
        public ContextMenu Menu
        {
            get
            {
                if (menu == null)
                {
                    menu = new ContextMenu();
                    MenuItem item = new MenuItem();
                    item.Header = "Group";
                    menu.Items.Add(item);
                    //groupMenuItem = 0;

                    item = new MenuItem();
                    item.Header = "Show Caption";
                    menu.Items.Add(item);
                    //showCaptionMenuItemm = 1;

                    item = new MenuItem();
                    item.Header = "Show AddNew";
                    menu.Items.Add(item);
                    //showCaptionMenuItemm = 2;

                    item = new MenuItem();
                    item.Header = "Show Summary";
                    menu.Items.Add(item);

                    item = new MenuItem();
                    item.Header = "Show RowHeader";
                    menu.Items.Add(item);

                    item = new MenuItem();
                    item.Header = "Show TopLevelGroup";
                    menu.Items.Add(item);

                    item = new MenuItem();
                    item.Header = "ExpandAll";
                    menu.Items.Add(item);

                    item = new MenuItem();
                    item.Header = "CollapseAll";
                    menu.Items.Add(item);

                    item = new MenuItem();
                    item.Header = "Clear Sort";
                    menu.Items.Add(item);

                    item = new MenuItem();
                    item.Header = "Clear All Groups";
                    menu.Items.Add(item);

                    menu.AddHandler(MenuItem.ClickEvent, new RoutedEventHandler(menuClick));
                }

                return menu;
            }
        }

        void menuClick(object sender, RoutedEventArgs e)
        {
            MenuItem item = e.Source as MenuItem;
            int menuHit = menu.Items.IndexOf(item);
            switch (menuHit)
            {
                case groupMenuItem:
                    if (rightClickedColumnName.Length > 0)
                    {
                        ToggleGroupColumn(rightClickedColumnName);
                    }
                    break;
                case showCaptionMenuItem:
                    ShowCaption = !ShowCaption;
                    ResetDisplay();
                    break;
                case showAddNewMenuItem:
                    ShowAddNewRecord = !ShowAddNewRecord;
                    break;
                case showSummaryMenuItem:
                    ShowSummary = !ShowSummary;
                    break;
                case showTopLevelGroupMenuItem:
                    {
                        bool b = IsTopLevelGroupHidden();
                        GridControl.RowHeights.SetHidden(rowTopLevelCaption, rowTopLevelCaption, !b);
                        break;
                    }
                case expandAllMenuItem:
                    ExpandAllGroups();
                    break;
                case collapseAllMenuItem:
                    CollapseAllGroups();
                    break;
                case showRowHeaderMenuItem:
                    {
                        int count = 0;
                        bool b = GridControl.ColumnWidths.GetHidden(rowHeaderColumn, out count);
                        GridControl.ColumnWidths.SetHidden(rowHeaderColumn, rowHeaderColumn, !b);
                        break;
                    }
                case clearSortMenuItem:
                    groupingEngine.TableDescriptor.SortedColumns.Clear();
                    ResetDisplay();
                    break;
                case clearAllGroupsMenuItem:
                    groupingEngine.TableDescriptor.GroupedColumns.Clear();
                    ResetDisplay();
                    break;
                default:
                    break;
            }
        }

        private string rightClickedColumnName = "";
        private DisplayRowType rightClickDisplayRowType = DisplayRowType.Unknown;
        void GridControl_PreviewMouseRightButtonDown(object sender, System.Windows.Input.MouseButtonEventArgs e)
        {
            //ignore rightmouse downs to avoid selections
            e.Handled = true;
        }

        private void EnableContextMenu()
        {
            GridControl.ContextMenu = Menu;

            //groupMenuItem
            if (rightClickedColumnName.Length > 0)
            {
                ((MenuItem)Menu.Items[groupMenuItem]).IsEnabled = true;
                bool isGrouped = groupingEngine.TableDescriptor.GroupedColumns.IndexOf(rightClickedColumnName) > -1;
                ((MenuItem)Menu.Items[groupMenuItem]).Header = isGrouped ? string.Format("Ungroup {0}", rightClickedColumnName)
                    : string.Format("Group {0}", rightClickedColumnName);
            }
            else
            {
                ((MenuItem)Menu.Items[groupMenuItem]).IsEnabled = false;
            }


            //showCaptionMenuItem
            ((MenuItem)Menu.Items[showCaptionMenuItem]).Header = ShowCaption ? "Hide Caption" : "Show Caption";
            ((MenuItem)Menu.Items[showAddNewMenuItem]).Header = ShowAddNewRecord ? "Hide AddNewRecord" : "Show AddNewRecord";
            ((MenuItem)Menu.Items[showSummaryMenuItem]).Header = ShowSummary ? "Hide Summary" : "Show Summary";

            int count = 0;
            bool b = GridControl.ColumnWidths.GetHidden(rowHeaderColumn, out count);
            ((MenuItem)Menu.Items[showRowHeaderMenuItem]).Header = b ? "Show RowHeader" : "Hide RowHeader";

            b = IsTopLevelGroupHidden();
            ((MenuItem)Menu.Items[showTopLevelGroupMenuItem]).Header = b ? "Show TopLevelGroup" : "Hide TopLevelGroup";

            ((MenuItem)Menu.Items[clearSortMenuItem]).IsEnabled = (GroupingEngine.TableDescriptor.SortedColumns.Count > 0);
            ((MenuItem)Menu.Items[clearAllGroupsMenuItem]).IsEnabled = (GroupingEngine.TableDescriptor.GroupedColumns.Count > 0);

        }

        void GridControl_PreviewMouseRightButtonUp(object sender, System.Windows.Input.MouseButtonEventArgs e)
        {


            GridControlBase grid = (GridControlBase)sender;
            RowColumnIndex cell = grid.PointToCellRowColumnIndex(e);
            Element el;
            rightClickDisplayRowType = GetDisplayRowType(cell.RowIndex, out el);
            if (rightClickDisplayRowType == DisplayRowType.Record ||
                rightClickDisplayRowType == DisplayRowType.LastRecord ||
                rightClickDisplayRowType == DisplayRowType.AddNewRecord &&
                cell.ColumnIndex > fixedColumns)
            {
                //this prevents the cell contextmenu
                e.Handled = true;
            }
            else
            {
                //expand collapse a group
                if (cell.ColumnIndex >= fixedColumns)
                {
                    rightClickedColumnName = ColumnIndexToName(cell.ColumnIndex);
                }
                else
                {
                    rightClickedColumnName = "";
                }
                EnableContextMenu();
            }
        }

        #endregion

        #region Properties

        private Brush spacerSpanInfoBrush = null;

        /// <summary>
        /// Gets or sets a brush that controls the CellSpanBackground of the spacer region under the Expand cell next to the records in a group.
        /// </summary>
        public Brush SpacerSpanInfoBrush
        {
            get { return spacerSpanInfoBrush; }
            set
            {
                if (spacerSpanInfoBrush != value)
                {
                    spacerSpanInfoBrush = value;
                    ResetDisplay();
                }
            }
        }

        private Brush recordSpanInfoBrush = null;

        /// <summary>
        /// Gets or sets a brush that controls the CellSpanBackground of the region under the Record cells in a group.
        /// </summary>
        public Brush RecordSpanInfoBrush
        {
            get { return recordSpanInfoBrush; }
            set
            {
                if (recordSpanInfoBrush != value)
                {
                    recordSpanInfoBrush = value;
                    ResetDisplay();
                }

            }
        }

        private bool allowClickSort = true;

        /// <summary>
        /// Enables or disables sorting by clicking a row header. 
        /// </summary>
        public bool AllowClickSort
        {
            get { return allowClickSort; }
            set { allowClickSort = value; }
        }


        private int expandIndentWidth = 10;

        /// <summary>
        /// Gets or sets the width of the expand column.
        /// </summary>
        public int ExpandIndentWidth
        {
            get { return expandIndentWidth; }
            set { expandIndentWidth = value; }
        }

        private bool allowGroupIndent = true;

        /// <summary>
        /// Gets or sets whether nested groups are indented.
        /// </summary>
        public bool AllowGroupIndent
        {
            get { return allowGroupIndent; }
            set
            {
                if (allowGroupIndent != value)
                {

                    allowGroupIndent = value;
                    if (allowGroupIndent)
                    {
                        GridControl.Model.ColumnWidths[expandColumn] = GetExpandColumnWidth();
                    }
                    else
                    {
                        GridControl.Model.ColumnWidths[expandColumn] = GetExpandColumnWidth();
                    }
                }
            }
        }

        private bool showAddNewRecord = false;

        /// <summary>
        /// Gets or sets whether an AddNew record is displayed in the GridControl.
        /// </summary>
        public bool ShowAddNewRecord
        {
            get { return showAddNewRecord; }
            set
            {
                if (showAddNewRecord != value)
                {
                    showAddNewRecord = value;
                    ApplyChange();
                }
            }
        }
        private bool showCaption = true;

        /// <summary>
        /// Gets or sets whether captions are displayed.
        /// </summary>
        public bool ShowCaption
        {
            get { return showCaption; }
            set
            {
                if (showCaption != value)
                {
                    showCaption = value;
                    ApplyChange();
                }
            }
        }

        private bool showSummary = false;

        /// <summary>
        /// Gets or sets whether a Summary row is dispalyed.
        /// </summary>
        public bool ShowSummary
        {
            get { return showSummary; }
            set
            {
                if (showSummary != value)
                {
                    showSummary = value;
                    ApplyChange();
                }
            }
        }

        private bool suspendApplyChange = false;
        /// <summary>
        /// Gets or sets whether the GridControl responds to property changes.
        /// </summary>
        public bool SuspendApplyChange
        {
            get { return suspendApplyChange; }
            set { suspendApplyChange = value; }
        }

        /// <summary>
        /// Forces the Grouping Engine to reload the data from its DataSource.
        /// </summary>
        public void ApplyChange()
        {
            if (!SuspendApplyChange)
            {
                this.GroupingEngine.Table.Reload();
                ResetDisplay();

            }
        }

        Engine groupingEngine = null;
        /// <summary>
        /// Returns the grouping Engine used in the helper class.
        /// </summary>
        public Engine GroupingEngine
        {
            get { return groupingEngine; }
        }

        GridControl gridControl = null;
        /// <summary>
        /// Returns the GridControl associated with this helper.
        /// </summary>
        public GridControl GridControl
        {
            get { return gridControl; }
        }

        List<string> visibleColumns = null;

        /// <summary>
        /// Gets or sets a list of columns to be displayed in the GridControl.
        /// </summary>
        public List<string> VisibleColumns
        {
            get
            {
                if (visibleColumns == null)
                {
                    visibleColumns = new List<string>();
                    foreach (FieldDescriptor fd in GroupingEngine.TableDescriptor.Fields)
                    {
                        visibleColumns.Add(fd.MappingName);
                    }
                }
                return visibleColumns;
            }
            set { visibleColumns = value; }
        }


        Dictionary<string, ColumnStyleInfo> columnStyles = null;

        /// <summary>
        /// Gets or sets a collection ColumnStyleInfo objects used in this GridControl.
        /// </summary>
        public Dictionary<string, ColumnStyleInfo> ColumnStyles
        {
            get
            {
                if (columnStyles == null)
                {
                    columnStyles = new Dictionary<string, ColumnStyleInfo>();
                }
                return columnStyles;
            }
            set { columnStyles = value; }
        }


        #endregion

        #region Element CellStyles

        GridStyleInfo cellStyleForColumnHeader = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of column header cells.
        /// </summary>
        public GridStyleInfo CellStyleForColumnHeader
        {
            get
            {
                if (cellStyleForColumnHeader == null)
                {
                    cellStyleForColumnHeader = new GridStyleInfo();
                    //set the defaults
                    cellStyleForColumnHeader.CellType = "SortHeader";
                    cellStyleForColumnHeader.HorizontalAlignment = HorizontalAlignment.Center;
                    cellStyleForColumnHeader.VerticalAlignment = VerticalAlignment.Center;
                    cellStyleForColumnHeader.TextWrapping = TextWrapping.Wrap;
                    cellStyleForColumnHeader.Font.FontWeight = FontWeights.Bold;
                }
                return cellStyleForColumnHeader;
            }
            set { cellStyleForColumnHeader = value; }
        }

        GridStyleInfo cellStyleForCaption = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of caption cells.
        /// </summary>
        public GridStyleInfo CellStyleForCaption
        {
            get
            {
                if (cellStyleForCaption == null)
                {
                    cellStyleForCaption = new GridStyleInfo();
                    cellStyleForCaption.CellType = "Static";
                    cellStyleForCaption.HorizontalAlignment = HorizontalAlignment.Left;
                    cellStyleForCaption.BorderMargins.Left = 5;
                }
                return cellStyleForCaption;
            }
            set { cellStyleForCaption = value; }
        }

        GridStyleInfo cellStyleForSummarySpacer = null;
        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the spacer cell next in summary rows.
        /// </summary>
        public GridStyleInfo CellStyleForSummarySpacer
        {
            get
            {
                if (cellStyleForSummarySpacer == null)
                {
                    cellStyleForSummarySpacer = new GridStyleInfo();
                    cellStyleForSummarySpacer.CellType = "Static";
                    cellStyleForSummarySpacer.Borders.Right = null;
                }
                return cellStyleForSummarySpacer;
            }
            set { cellStyleForSummarySpacer = value; }
        }

        GridStyleInfo cellStyleForSummary = null;
        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of summary cells.
        /// </summary>
        public GridStyleInfo CellStyleForSummary
        {
            get
            {
                if (cellStyleForSummary == null)
                {
                    cellStyleForSummary = new GridStyleInfo();
                    cellStyleForSummary.CellType = "Static";
                    cellStyleForSummary.Borders.Right = null;
                }
                return cellStyleForSummary;
            }
            set { cellStyleForSummary = value; }
        }
        GridStyleInfo cellStyleForExpand = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of expand cells.
        /// </summary>
        public GridStyleInfo CellStyleForExpand
        {
            get
            {
                if (cellStyleForExpand == null)
                {
                    cellStyleForExpand = new GridStyleInfo();
                    cellStyleForExpand.Background = Brushes.AliceBlue;
                    cellStyleForExpand.CellType = "Static";
                    cellStyleForExpand.BorderMargins.Left = 5;
                    //cellStyleForExpand.BorderMargins.Top = 0;
                    cellStyleForExpand.Font.FontWeight = FontWeights.Bold;
                    cellStyleForExpand.Font.FontSize = 14;
                    cellStyleForExpand.VerticalAlignment = VerticalAlignment.Top;
                    cellStyleForExpand.HorizontalAlignment = HorizontalAlignment.Left;
                }
                return cellStyleForExpand;
            }
            set { cellStyleForExpand = value; }
        }
        GridStyleInfo cellStyleForSpacer = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the spacer cell next to record rows.
        /// </summary>
        public GridStyleInfo CellStyleForSpacer
        {
            get
            {
                if (cellStyleForSpacer == null)
                {
                    cellStyleForSpacer = new GridStyleInfo();
                    cellStyleForSpacer.Borders.Top = null;
                    cellStyleForSpacer.Borders.Bottom = null;
                    cellStyleForSpacer.Enabled = false;
                }
                return cellStyleForSpacer;
            }
            set { cellStyleForSpacer = value; }
        }

        GridStyleInfo cellStyleForColumnHeaderSpacer = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the spacer cell next to the column header.
        /// </summary>
        public GridStyleInfo CellStyleForColumnHeaderSpacer
        {
            get
            {
                if (cellStyleForColumnHeaderSpacer == null)
                {
                    cellStyleForColumnHeaderSpacer = new GridStyleInfo();
                    cellStyleForColumnHeaderSpacer.Borders.Top = null;
                    cellStyleForColumnHeaderSpacer.Borders.Bottom = null;
                    cellStyleForColumnHeaderSpacer.Background = Brushes.Pink;
                }
                return cellStyleForColumnHeaderSpacer;
            }
            set { cellStyleForColumnHeaderSpacer = value; }
        }

        GridStyleInfo cellStyleForLastSpacer = null;
        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the spacer cell next to the last record row.
        /// </summary>
        public GridStyleInfo CellStyleForLastSpacer
        {
            get
            {
                if (cellStyleForLastSpacer == null)
                {
                    cellStyleForLastSpacer = new GridStyleInfo();
                    cellStyleForLastSpacer.Borders.Top = null;
                    cellStyleForLastSpacer.Enabled = false;
                }
                return cellStyleForLastSpacer;
            }
            set { cellStyleForLastSpacer = value; }
        }

        GridStyleInfo cellStyleForRecordRowHeader = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the header cells on record rows.
        /// </summary>
        public GridStyleInfo CellStyleForRecordRowHeader
        {
            get
            {
                if (cellStyleForRecordRowHeader == null)
                {
                    cellStyleForRecordRowHeader = new GridStyleInfo();
                }
                return cellStyleForRecordRowHeader;
            }
            set { cellStyleForRecordRowHeader = value; }
        }
        GridStyleInfo cellStyleForLastRecordRowHeader = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the header cell on the last record row.
        /// </summary>
        public GridStyleInfo CellStyleForLastRecordRowHeader
        {
            get
            {
                if (cellStyleForLastRecordRowHeader == null)
                {
                    cellStyleForLastRecordRowHeader = new GridStyleInfo();
                    cellStyleForLastRecordRowHeader.Enabled = false;
                }
                return cellStyleForLastRecordRowHeader;
            }
            set { cellStyleForLastRecordRowHeader = value; }
        }
        GridStyleInfo cellStyleForColumnHeaderRowHeader = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the header cell on the column header row.
        /// </summary>
        public GridStyleInfo CellStyleForColumnHeaderRowHeader
        {
            get
            {
                if (cellStyleForColumnHeaderRowHeader == null)
                {
                    cellStyleForColumnHeaderRowHeader = new GridStyleInfo();
                }
                return cellStyleForColumnHeaderRowHeader;
            }
            set { cellStyleForColumnHeaderRowHeader = value; }
        }
        GridStyleInfo cellStyleForCaptionRowHeader = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the header cell on caption rows.
        /// </summary>
        public GridStyleInfo CellStyleForCaptionRowHeader
        {
            get
            {
                if (cellStyleForCaptionRowHeader == null)
                {
                    cellStyleForCaptionRowHeader = new GridStyleInfo();
                }
                return cellStyleForCaptionRowHeader;
            }
            set { cellStyleForCaptionRowHeader = value; }
        }
        GridStyleInfo cellStyleForSummaryRowHeader = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the header cell on summary rows.
        /// </summary>
        public GridStyleInfo CellStyleForSummaryRowHeader
        {
            get
            {
                if (cellStyleForSummaryRowHeader == null)
                {
                    cellStyleForSummaryRowHeader = new GridStyleInfo();
                }
                return cellStyleForSummaryRowHeader;
            }
            set { cellStyleForSummaryRowHeader = value; }
        }
        GridStyleInfo cellStyleForAddNewRowHeader = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object that controls the style properties of the header cell on AddNew rows.
        /// </summary>
        public GridStyleInfo CellStyleForAddNewRowHeader
        {
            get
            {
                if (cellStyleForAddNewRowHeader == null)
                {
                    cellStyleForAddNewRowHeader = new GridStyleInfo();
                }
                return cellStyleForAddNewRowHeader;
            }
            set { cellStyleForAddNewRowHeader = value; }
        }

        #endregion

    }

    #region ColumnStyleInfo class

/// <summary>
/// Holds properties associated with columns in the GridControl.
/// </summary>
    public class ColumnStyleInfo
    {
        GridStyleInfo style = null;

        /// <summary>
        /// Gets of sets the GridStyleInfo object for the column.
        /// </summary>
        public GridStyleInfo Style
        {
            get
            {
                if (style == null)
                    style = new GridStyleInfo();
                return style; 
            }
            set { style = value; }
        }
       
        string name = "";

        /// <summary>
        /// Gets of sets the mapping name of the column.
        /// </summary>
        public string Name
        {
            get { return name; }
            set { name = value; }
        }
        string headerText = "";

        /// <summary>
        /// Gets of sets the HeaderText of the column.
        /// </summary>
        public string HeaderText
        {
            get 
            {
                if (headerText.Length == 0)
                    return Name;
                return headerText; 
            }
            set { headerText = value; }
        }
    }
    #endregion

    #region enums
    /// <summary>
    /// The grouping element type of object in a GridControl row.
    /// </summary>
    public enum DisplayRowType
    {
        ColumnHeader,
        Caption,
        Record,
        LastRecord,
        AddNewRecord,
        SummarySection,
        Unknown
    }
    #endregion

    #region Cell Renderers

    public class SortHeaderCellRenderer : GridVirtualizingCellRenderer<StackPanel>
    {
        GridControl grid;
        GroupingHelper helper;
        BitmapImage up;
        BitmapImage down;
        private int arrowWidth = 0;
        TextBlock textBlock;
        Image c;

        public void InitializeRenderer(GridControl grid, GroupingHelper helper)
        {
            this.grid = grid;
            this.helper = helper;
            Uri uri = new Uri("pack://application:,,/images/Down.ico");
            down = new BitmapImage(uri);
            uri = new Uri("pack://application:,,/images/upIcon.ico");
            up = new BitmapImage(uri);
            arrowWidth = 13;
        }

        public override void OnInitializeContent(StackPanel uiElement, GridRenderStyleInfo cellInfo)
        {
            // Console.WriteLine("OnInitializeContent");

            uiElement.Children.Clear();

            uiElement.Orientation = Orientation.Horizontal;

            textBlock = new TextBlock();
            textBlock.Text = cellInfo.Text;
            textBlock.TextAlignment = TextAlignment.Center;

            int colIndex = cellInfo.CellIdentity.CellRowColumnIndex.ColumnIndex;
            textBlock.Width = grid.ColumnWidths[colIndex] - arrowWidth;

            textBlock.VerticalAlignment = cellInfo.VerticalAlignment;
            textBlock.TextWrapping = cellInfo.TextWrapping;
            textBlock.FontWeight = cellInfo.Font.FontWeight;
  
            uiElement.Children.Add(textBlock);

            string name = helper.ColumnIndexToName(colIndex);
             int i = -1;
            if ((i = helper.GroupingEngine.TableDescriptor.SortedColumns.IndexOf(name)) > -1)
            {
                SortColumnDescriptor scd = helper.GroupingEngine.TableDescriptor.SortedColumns[i];
                // Image 
                c = new Image();
                if (scd.SortDirection == ListSortDirection.Ascending)
                {
                    c.Source = up;
                }
                else //if (scd.SortDirection == ListSortDirection.Descending)
                {
                    c.Source = down;
                }
                uiElement.Children.Add(c);
            }

        }
    }

    #endregion

    #region custom grouping engine to limit display of AddNewRecords, CaptionRows and SummarySections

    public class LimitedEngine : Engine
    {
        GroupingHelper helper = null;
        public LimitedEngine(GroupingHelper helper)
            : base()
        {
            this.helper = helper;
        }

        //public override Table CreateTable(TableDescriptor tableDescriptor, Table parentTable)
        //{
        //    return new LimitedTable(tableDescriptor, parentTable, helper);
        //}
        public override Group CreateGroup(Section parent)
        {
            return new LimitedGroup(parent, helper);
        }
        public override ChildTable CreateChildTable(Element parent)
        {
            return new LimitedChildTable(parent, helper);
        }
        
    }


    public class LimitedChildTable : ChildTable
    {
        GroupingHelper helper;
        public LimitedChildTable(Element parent, GroupingHelper helper)
            : base(parent)
        {
            this.helper = helper;
        }
        public override bool IsChildVisible(Element el)
        {
            if (!helper.ShowAddNewRecord && el is AddNewRecordSection)
                return false;
            if (!helper.ShowCaption && el is CaptionSection)
                return false;
            if (!helper.ShowSummary && el is SummarySection)
                return false;

            return base.IsChildVisible(el);
        }
    }
    //public class LimitedTable : Table
    //{
    //    GroupingHelper helper;

    //    public LimitedTable(TableDescriptor tableDescriptor, Table parentTable, GroupingHelper helper)
    //        : base(tableDescriptor, parentTable)
    //    {
    //        this.helper = helper;
    //    }

    //    public override bool IsChildVisible(Element el)
    //    {
    //        return base.IsChildVisible(el);
    //    }
        
    //}

    public class LimitedGroup : Group
    {
        GroupingHelper helper;

        public LimitedGroup(Section parent, GroupingHelper helper)
            : base(parent)
        {
            this.helper = helper;
        }

        protected override void OnInitializeSections(bool hasRecords, SortColumnDescriptorCollection sortColumns)
        {
         //   base.OnInitializeSections(hasRecords, sortColumns);
            this.Sections.Add(this.ParentTableDescriptor.CreateCaptionSection(this));
            if (hasRecords)
                this.Sections.Add(this.ParentTableDescriptor.CreateRecordsDetails(this, sortColumns));
            else
                this.Sections.Add(this.ParentTableDescriptor.CreateGroupsDetails(this, sortColumns));
            if (helper.ShowSummary)
            {
                this.Sections.Add(this.ParentTableDescriptor.CreateSummarySection(this));
            }
        }
    }


    #endregion

    #region debug stuff

    public class Metrics
    {
 
        /// <summary>
        /// if true, Console.WriteLine is displayed.
        /// </summary>
        public static bool Enabled = false;
 
        
        
        public static DateTime startTime;

        public static void StartTimer()
        {
            startTime = DateTime.Now;
        }
        public static void DisplayElapsedTime(string note)
        {
            if (Enabled)
            {
                Console.WriteLine("{0:000.0000} {1}", ((TimeSpan)(DateTime.Now - startTime)).TotalSeconds, note);
                startTime = DateTime.Now;
            }
        }

        public static void DisplayNote(string note)
        {
            if (Enabled)
            {
                Console.WriteLine(note);
            }
        }

    }
    #endregion

}
